//======================================================================
// 
// Undelete.c
//
// Copyright (C) 1996-1998 Mark Russinovich and Bryce Cogswell
//
// Preserves files that are deleted by using the built-in Recycle-Bin
// functionality of NT 4.0.
//
//======================================================================
#include "ntddk.h"
#include "undelete.h"
#include "ioctlcmd.h"


//----------------------------------------------------------------------
//                         FORWARD DEFINES
//---------------------------------------------------------------------- 

//
// Undocumented ntoskrnl variable
//
extern PSHORT           NtBuildNumber;


NTSTATUS UndeleteDispatch( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp );

//
// These are prototypes for Undelete's Fast I/O hooks. The originals 
// prototypes can be found in NTDDK.H
// 
BOOLEAN  UndeleteFastIoCheckifPossible( IN PFILE_OBJECT FileObject, IN PLARGE_INTEGER FileOffset, 
                                        IN ULONG Length, IN BOOLEAN Wait, IN ULONG LockKey, IN BOOLEAN CheckForReadOperation,
                                        OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoRead( IN PFILE_OBJECT FileObject, IN PLARGE_INTEGER FileOffset, 
                             IN ULONG Length, IN BOOLEAN Wait, IN ULONG LockKey, OUT PVOID Buffer,
                             OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoWrite( IN PFILE_OBJECT FileObject, IN PLARGE_INTEGER FileOffset, 
                              IN ULONG Length, IN BOOLEAN Wait, IN ULONG LockKey, IN PVOID Buffer,
                              OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoQueryBasicInfo( IN PFILE_OBJECT FileObject, IN BOOLEAN Wait, 
                                       OUT PFILE_BASIC_INFORMATION Buffer,
                                       OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoQueryStandardInfo( IN PFILE_OBJECT FileObject, IN BOOLEAN Wait, 
                                          OUT PFILE_STANDARD_INFORMATION Buffer,
                                          OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoLock( IN PFILE_OBJECT FileObject, IN PLARGE_INTEGER FileOffset,
                             IN PLARGE_INTEGER Length, PEPROCESS ProcessId, ULONG Key,
                             BOOLEAN FailImmediately, BOOLEAN ExclusiveLock,
                             OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoUnlockSingle( IN PFILE_OBJECT FileObject, IN PLARGE_INTEGER FileOffset,
                                     IN PLARGE_INTEGER Length, PEPROCESS ProcessId, ULONG Key,
                                     OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoUnlockAll( IN PFILE_OBJECT FileObject, PEPROCESS ProcessId,
                                  OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoUnlockAllByKey( IN PFILE_OBJECT FileObject, PEPROCESS ProcessId, ULONG Key,
                                       OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoDeviceControl( IN PFILE_OBJECT FileObject, IN BOOLEAN Wait,
                                      IN PVOID InputBuffer, IN ULONG InputBufferLength, 
                                      OUT PVOID OutbufBuffer, IN ULONG OutputBufferLength, IN ULONG IoControlCode,
                                      OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject );
VOID     UndeleteFastIoAcquireFile( PFILE_OBJECT FileObject );
VOID     UndeleteFastIoReleaseFile( PFILE_OBJECT FileObject );
VOID     UndeleteFastIoDetachDevice( PDEVICE_OBJECT SourceDevice, PDEVICE_OBJECT TargetDevice );

//
// These are new NT 4.0 Fast I/O calls
//
BOOLEAN  UndeleteFastIoQueryNetworkOpenInfo(IN PFILE_OBJECT FileObject,
                                            IN BOOLEAN Wait, OUT struct _FILE_NETWORK_OPEN_INFORMATION *Buffer,
                                            OUT struct _IO_STATUS_BLOCK *IoStatus, IN PDEVICE_OBJECT DeviceObject );
NTSTATUS UndeleteFastIoAcquireForModWrite( IN PFILE_OBJECT FileObject,
                                           IN PLARGE_INTEGER EndingOffset, OUT struct _ERESOURCE **ResourceToRelease,
                                           IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoMdlRead( IN PFILE_OBJECT FileObject,
                                IN PLARGE_INTEGER FileOffset, IN ULONG Length,
                                IN ULONG LockKey, OUT PMDL *MdlChain, OUT PIO_STATUS_BLOCK IoStatus,
                                IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoMdlReadComplete( IN PFILE_OBJECT FileObject,
                                        IN PMDL MdlChain, IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoPrepareMdlWrite( IN PFILE_OBJECT FileObject,
                                        IN PLARGE_INTEGER FileOffset, IN ULONG Length, IN ULONG LockKey,
                                        OUT PMDL *MdlChain, OUT PIO_STATUS_BLOCK IoStatus,
                                        IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoMdlWriteComplete( IN PFILE_OBJECT FileObject,
                                         IN PLARGE_INTEGER FileOffset, IN PMDL MdlChain,
                                         IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoReadCompressed( IN PFILE_OBJECT FileObject,
                                       IN PLARGE_INTEGER FileOffset, IN ULONG Length,
                                       IN ULONG LockKey, OUT PVOID Buffer, OUT PMDL *MdlChain,
                                       OUT PIO_STATUS_BLOCK IoStatus,
                                       OUT struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
                                       IN ULONG CompressedDataInfoLength, IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoWriteCompressed( IN PFILE_OBJECT FileObject,
                                        IN PLARGE_INTEGER FileOffset, IN ULONG Length,
                                        IN ULONG LockKey, IN PVOID Buffer, OUT PMDL *MdlChain,
                                        OUT PIO_STATUS_BLOCK IoStatus,
                                        IN struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
                                        IN ULONG CompressedDataInfoLength, IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoMdlReadCompleteCompressed( IN PFILE_OBJECT FileObject,
                                                  IN PMDL MdlChain, IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoMdlWriteCompleteCompressed( IN PFILE_OBJECT FileObject,
                                                   IN PLARGE_INTEGER FileOffset, IN PMDL MdlChain,
                                                   IN PDEVICE_OBJECT DeviceObject );
BOOLEAN  UndeleteFastIoQueryOpen( IN struct _IRP *Irp,
                                  OUT PFILE_NETWORK_OPEN_INFORMATION NetworkInformation,
                                  IN PDEVICE_OBJECT DeviceObject );
NTSTATUS UndeleteFastIoReleaseForModWrite( IN PFILE_OBJECT FileObject,
                                           IN struct _ERESOURCE *ResourceToRelease, IN PDEVICE_OBJECT DeviceObject );
NTSTATUS UndeleteFastIoAcquireForCcFlush( IN PFILE_OBJECT FileObject,
                                          IN PDEVICE_OBJECT DeviceObject );
NTSTATUS UndeleteFastIoReleaseForCcFlush( IN PFILE_OBJECT FileObject,
                                          IN PDEVICE_OBJECT DeviceObject );

//----------------------------------------------------------------------
//                         GLOBALS
//---------------------------------------------------------------------- 

//
// This is Undelete's user-inteface device object. It is addressed
// by calls from the GUI including CreateFile and DeviceIoControl
//
PDEVICE_OBJECT      GUIDevice;

//
// Mutex that is shared with GUI to synchronize access to the
// undelete bins
//
PKMUTEX            UndeleteMutex = NULL;

//
// Table of our hook devices for each drive letter. This makes it
// easy to look up the device object that was created to hook a 
// particular drive.
//
PDEVICE_OBJECT      LDriveDevices[26];

//
// System process
// 
PEPROCESS           SystemProcess;

//
// Extensions that should not be undeleted
//
ULONG               NumFilters = 0;
PFILTERENTRY        FilterList = NULL;

//
// Directories that should not be protected
//
ULONG               SizeDirFilters = 0;
ULONG               NumDirFilters = 0;
PCHAR               *DirFilterArray;
CHAR                *DirFilterBuffer;

//
// This Undelete's Fast I/O dispatch table. Note that NT assumes that
// file system drivers support some Fast I/O calls, so this table must
// be present for an file system filter driver
//
FAST_IO_DISPATCH    FastIOHook = {
    sizeof(FAST_IO_DISPATCH), 
    UndeleteFastIoCheckifPossible,
    UndeleteFastIoRead,
    UndeleteFastIoWrite,
    UndeleteFastIoQueryBasicInfo,
    UndeleteFastIoQueryStandardInfo,
    UndeleteFastIoLock,
    UndeleteFastIoUnlockSingle,
    UndeleteFastIoUnlockAll,
    UndeleteFastIoUnlockAllByKey,
    UndeleteFastIoDeviceControl,
    UndeleteFastIoAcquireFile,
    UndeleteFastIoReleaseFile,
    UndeleteFastIoDetachDevice,

    //
    // new for NT 4.0
    //
    UndeleteFastIoQueryNetworkOpenInfo,
    UndeleteFastIoAcquireForModWrite,
    UndeleteFastIoMdlRead,
    UndeleteFastIoMdlReadComplete,
    UndeleteFastIoPrepareMdlWrite,
    UndeleteFastIoMdlWriteComplete,
    UndeleteFastIoReadCompressed,
    UndeleteFastIoWriteCompressed,
    UndeleteFastIoMdlReadCompleteCompressed,
    UndeleteFastIoMdlWriteCompleteCompressed,
    UndeleteFastIoQueryOpen,
    UndeleteFastIoReleaseForModWrite,
    UndeleteFastIoAcquireForCcFlush,
    UndeleteFastIoReleaseForCcFlush
};


//----------------------------------------------------------------------
//                F A S T I O   R O U T I N E S
//----------------------------------------------------------------------

//----------------------------------------------------------------------
//
// UndeleteFastIoCheckIfPossible
//
//----------------------------------------------------------------------
BOOLEAN  UndeleteFastIoCheckifPossible( IN PFILE_OBJECT FileObject, IN PLARGE_INTEGER FileOffset, 
                                        IN ULONG Length, IN BOOLEAN Wait, IN ULONG LockKey, IN BOOLEAN CheckForReadOperation,
                                        OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject ) {
    BOOLEAN         retval = FALSE;
    PHOOK_EXTENSION hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoCheckIfPossible ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoCheckIfPossible( 
            FileObject, FileOffset, Length,
            Wait, LockKey, CheckForReadOperation, IoStatus, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
// 
// UndeleteFastIoRead
//
//----------------------------------------------------------------------
BOOLEAN  UndeleteFastIoRead( IN PFILE_OBJECT FileObject, IN PLARGE_INTEGER FileOffset, 
                             IN ULONG Length, IN BOOLEAN Wait, IN ULONG LockKey, OUT PVOID Buffer,
                             OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject ) {
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoRead ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoRead( 
            FileObject, FileOffset, Length,
            Wait, LockKey, Buffer, IoStatus, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoWrite
//
//----------------------------------------------------------------------
BOOLEAN  UndeleteFastIoWrite( IN PFILE_OBJECT FileObject, IN PLARGE_INTEGER FileOffset, 
                              IN ULONG Length, IN BOOLEAN Wait, IN ULONG LockKey, IN PVOID Buffer,
                              OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject ) {
    BOOLEAN              retval = FALSE;
    PHOOK_EXTENSION      hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoWrite ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoWrite( 
            FileObject, FileOffset, Length, Wait, LockKey, 
            Buffer, IoStatus, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoQueryBasicinfo
//
//----------------------------------------------------------------------
BOOLEAN  UndeleteFastIoQueryBasicInfo( IN PFILE_OBJECT FileObject, IN BOOLEAN Wait, 
                                       OUT PFILE_BASIC_INFORMATION Buffer,
                                       OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject ) {
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoQueryBasicInfo ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoQueryBasicInfo( 
            FileObject, Wait, Buffer, IoStatus, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoQueryStandardInfo
//
//----------------------------------------------------------------------
BOOLEAN  UndeleteFastIoQueryStandardInfo( IN PFILE_OBJECT FileObject, IN BOOLEAN Wait, 
                                          OUT PFILE_STANDARD_INFORMATION Buffer,
                                          OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject ) {
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoQueryStandardInfo ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoQueryStandardInfo( 
            FileObject, Wait, Buffer, IoStatus, hookExt->FileSystem );

    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoLock
//
//----------------------------------------------------------------------
BOOLEAN  UndeleteFastIoLock( IN PFILE_OBJECT FileObject, IN PLARGE_INTEGER FileOffset,
                             IN PLARGE_INTEGER Length, PEPROCESS ProcessId, ULONG Key,
                             BOOLEAN FailImmediately, BOOLEAN ExclusiveLock,
                             OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject ) {
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;   

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoLock ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoLock( 
            FileObject, FileOffset, Length, ProcessId, Key, FailImmediately, 
            ExclusiveLock, IoStatus, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoUnlockSingle
//
//----------------------------------------------------------------------
BOOLEAN  UndeleteFastIoUnlockSingle( IN PFILE_OBJECT FileObject, IN PLARGE_INTEGER FileOffset,
                                     IN PLARGE_INTEGER Length, PEPROCESS ProcessId, ULONG Key,
                                     OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject ) {
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;   
    
    if( hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoUnlockSingle ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoUnlockSingle(
            FileObject, FileOffset, Length, ProcessId, Key, 
            IoStatus, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoUnlockAll
//
//----------------------------------------------------------------------
BOOLEAN  UndeleteFastIoUnlockAll( IN PFILE_OBJECT FileObject, PEPROCESS ProcessId,
                                  OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject ) {
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if ( hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoUnlockAll ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoUnlockAll( 
            FileObject, ProcessId, IoStatus, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoUnlockAllByKey
//
//----------------------------------------------------------------------    
BOOLEAN  UndeleteFastIoUnlockAllByKey( IN PFILE_OBJECT FileObject, PEPROCESS ProcessId, ULONG Key,
                                       OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject ) {
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoUnlockAllByKey ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoUnlockAllByKey( 
            FileObject, ProcessId, Key, IoStatus, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoQueryNetworkOpenInfo
//
//----------------------------------------------------------------------    
BOOLEAN UndeleteFastIoQueryNetworkOpenInfo(IN PFILE_OBJECT FileObject,
                                           IN BOOLEAN Wait,
                                           OUT struct _FILE_NETWORK_OPEN_INFORMATION *Buffer,
                                           OUT PIO_STATUS_BLOCK IoStatus,
                                           IN PDEVICE_OBJECT DeviceObject )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoQueryNetworkOpenInfo ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoQueryNetworkOpenInfo( 
            FileObject, Wait, Buffer, IoStatus, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoAcquireForModWrite
//
//----------------------------------------------------------------------    
NTSTATUS UndeleteFastIoAcquireForModWrite( IN PFILE_OBJECT FileObject,
                                           IN PLARGE_INTEGER EndingOffset,
                                           OUT struct _ERESOURCE **ResourceToRelease,
                                           IN PDEVICE_OBJECT DeviceObject )
{
    NTSTATUS            retval = STATUS_NOT_SUPPORTED;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->AcquireForModWrite ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->AcquireForModWrite( 
            FileObject, EndingOffset, ResourceToRelease, hookExt->FileSystem );

    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoMdlRead
//
//----------------------------------------------------------------------    
BOOLEAN UndeleteFastIoMdlRead( IN PFILE_OBJECT FileObject,
                               IN PLARGE_INTEGER FileOffset, IN ULONG Length,
                               IN ULONG LockKey, OUT PMDL *MdlChain,
                               OUT PIO_STATUS_BLOCK IoStatus,
                               IN PDEVICE_OBJECT DeviceObject )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->MdlRead ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->MdlRead( 
            FileObject, FileOffset, Length, LockKey, MdlChain, 
            IoStatus, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoMdlReadComplete
//
//----------------------------------------------------------------------    
BOOLEAN UndeleteFastIoMdlReadComplete( IN PFILE_OBJECT FileObject,
                                       IN PMDL MdlChain, IN PDEVICE_OBJECT DeviceObject )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->MdlReadComplete ) {

        retval = (BOOLEAN) hookExt->FileSystem->DriverObject->FastIoDispatch->MdlReadComplete( FileObject, 
                                                                                               MdlChain, hookExt->FileSystem );

    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoPrepareMdlWrite
//
//----------------------------------------------------------------------    
BOOLEAN UndeleteFastIoPrepareMdlWrite( IN PFILE_OBJECT FileObject,
                                       IN PLARGE_INTEGER FileOffset, IN ULONG Length,
                                       IN ULONG LockKey, OUT PMDL *MdlChain,
                                       OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;
    IoStatus->Status      = STATUS_NOT_SUPPORTED;
    IoStatus->Information = 0;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->PrepareMdlWrite ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->PrepareMdlWrite( 
            FileObject, FileOffset, Length, LockKey, MdlChain, IoStatus, 
            hookExt->FileSystem );
    } 
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoMdlWriteComplete
//
//----------------------------------------------------------------------    
BOOLEAN UndeleteFastIoMdlWriteComplete( IN PFILE_OBJECT FileObject,
                                        IN PLARGE_INTEGER FileOffset, IN PMDL MdlChain, 
                                        IN PDEVICE_OBJECT DeviceObject )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->MdlWriteComplete ) { 

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->MdlWriteComplete( 
            FileObject, FileOffset, MdlChain, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoReadCompressed
//
//----------------------------------------------------------------------    
BOOLEAN UndeleteFastIoReadCompressed( IN PFILE_OBJECT FileObject,
                                      IN PLARGE_INTEGER FileOffset, IN ULONG Length,
                                      IN ULONG LockKey, OUT PVOID Buffer,
                                      OUT PMDL *MdlChain, OUT PIO_STATUS_BLOCK IoStatus,
                                      OUT struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
                                      IN ULONG CompressedDataInfoLength, IN PDEVICE_OBJECT DeviceObject )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoReadCompressed ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoReadCompressed( 
            FileObject, FileOffset, Length, LockKey, Buffer, MdlChain, IoStatus,
            CompressedDataInfo, CompressedDataInfoLength, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoWriteCompressed
//
//----------------------------------------------------------------------    
BOOLEAN UndeleteFastIoWriteCompressed( IN PFILE_OBJECT FileObject,
                                       IN PLARGE_INTEGER FileOffset, IN ULONG Length,
                                       IN ULONG LockKey, OUT PVOID Buffer,
                                       OUT PMDL *MdlChain, OUT PIO_STATUS_BLOCK IoStatus,
                                       OUT struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
                                       IN ULONG CompressedDataInfoLength, IN PDEVICE_OBJECT DeviceObject )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoWriteCompressed ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoWriteCompressed( 
            FileObject, FileOffset, Length, LockKey, Buffer, MdlChain, IoStatus,
            CompressedDataInfo, CompressedDataInfoLength, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoMdlReadCompleteCompressed
//
//----------------------------------------------------------------------    
BOOLEAN UndeleteFastIoMdlReadCompleteCompressed( IN PFILE_OBJECT FileObject,
                                                 IN PMDL MdlChain, IN PDEVICE_OBJECT DeviceObject )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->MdlReadCompleteCompressed ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->MdlReadCompleteCompressed( 
            FileObject, MdlChain, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoMdlWriteCompleteCompressed
//
//----------------------------------------------------------------------    
BOOLEAN UndeleteFastIoMdlWriteCompleteCompressed( IN PFILE_OBJECT FileObject,
                                                  IN PLARGE_INTEGER FileOffset, IN PMDL MdlChain, 
                                                  IN PDEVICE_OBJECT DeviceObject )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension; 

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->MdlWriteCompleteCompressed ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->MdlWriteCompleteCompressed( 
            FileObject, FileOffset, MdlChain, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoQueryOpen
//
// This call actually passes an IRP! 
//
//----------------------------------------------------------------------    
BOOLEAN UndeleteFastIoQueryOpen( IN PIRP Irp,
                                 OUT PFILE_NETWORK_OPEN_INFORMATION NetworkInformation,
                                 IN PDEVICE_OBJECT DeviceObject )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    PFILE_OBJECT        fileobject;
    PIO_STACK_LOCATION  currentIrpStack;
    PIO_STACK_LOCATION  nextIrpStack;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoQueryOpen ) {

        currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
        nextIrpStack    = IoGetNextIrpStackLocation(Irp);
        fileobject      = currentIrpStack->FileObject;

        //
        // copy parameters down to next level in the stack
        //
        *nextIrpStack = *currentIrpStack;
        nextIrpStack->DeviceObject = hookExt->FileSystem;
        IoSetNextIrpStackLocation( Irp );

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoQueryOpen( 
            Irp, NetworkInformation, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoReleaseForModWrite
//
//----------------------------------------------------------------------    
NTSTATUS UndeleteFastIoReleaseForModWrite( IN PFILE_OBJECT FileObject,
                                           IN struct _ERESOURCE *ResourceToRelease,
                                           IN PDEVICE_OBJECT DeviceObject )
{
    NTSTATUS            retval = STATUS_NOT_SUPPORTED;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->ReleaseForModWrite ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->ReleaseForModWrite( 
            FileObject,  ResourceToRelease, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoAcquireForCcFlush
//
//----------------------------------------------------------------------    
NTSTATUS UndeleteFastIoAcquireForCcFlush( IN PFILE_OBJECT FileObject,
                                          IN PDEVICE_OBJECT DeviceObject )
{
    NTSTATUS            retval = STATUS_NOT_SUPPORTED;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->AcquireForCcFlush ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->AcquireForCcFlush( 
            FileObject, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoReleaseForCcFlush
//
//----------------------------------------------------------------------    
NTSTATUS UndeleteFastIoReleaseForCcFlush( IN PFILE_OBJECT FileObject,
                                          IN PDEVICE_OBJECT DeviceObject )
{
    NTSTATUS            retval = STATUS_NOT_SUPPORTED;
    PHOOK_EXTENSION     hookExt;

    hookExt = DeviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->ReleaseForCcFlush ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->ReleaseForCcFlush( 
            FileObject, hookExt->FileSystem );
    }
    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoDeviceControl
//
//----------------------------------------------------------------------
BOOLEAN  UndeleteFastIoDeviceControl( IN PFILE_OBJECT FileObject, IN BOOLEAN Wait,
                                      IN PVOID InputBuffer, IN ULONG InputBufferLength, 
                                      OUT PVOID OutputBuffer, IN ULONG OutputBufferLength, IN ULONG IoControlCode,
                                      OUT PIO_STATUS_BLOCK IoStatus, IN PDEVICE_OBJECT DeviceObject ) {
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    PCHAR               curDirPtr;
    NTSTATUS            ntStatus;
    ULONG               i;
    HANDLE              mutexHandle;

    if ( DeviceObject == GUIDevice )  {

        //
        // Its a message from our GUI!
        //
        IoStatus->Status      = STATUS_SUCCESS; 
        IoStatus->Information = 0;      

        switch ( IoControlCode ) {

        case UNDELETE_FILTER_UPDATE:

            //
            // Update of extension filters
            //
            if( FilterList ) ExFreePool( FilterList );

            NumFilters = InputBufferLength/sizeof(FILTERENTRY);
            FilterList = ExAllocatePool( NonPagedPool, InputBufferLength );
            memcpy( FilterList, InputBuffer, InputBufferLength );
            DbgPrint(("Undelete: %d filters in filter update\n", NumFilters ));
            break;

        case UNDELETE_DIR_FILTER_UPDATE:

            //
            // Update of directory filters
            //
            if( DirFilterArray ) ExFreePool( DirFilterArray );
            if( DirFilterBuffer) ExFreePool( DirFilterBuffer );

            SizeDirFilters = InputBufferLength;
            DirFilterBuffer = (PCHAR) ExAllocatePool( NonPagedPool, InputBufferLength );
            memcpy( DirFilterBuffer, InputBuffer, InputBufferLength );

            //
            // First, determine how many filters are here
            //
            curDirPtr = DirFilterBuffer;
            NumDirFilters = 0;
            while( curDirPtr < DirFilterBuffer + InputBufferLength ) {
	      
                NumDirFilters++;
                curDirPtr += strlen( curDirPtr )+1;
            }
	    
            //
            // Allocate the array and fill it in
            //
            DirFilterArray = (PCHAR *) ExAllocatePool( NonPagedPool, NumDirFilters );
            curDirPtr = DirFilterBuffer;
            for( i = 0; i < NumDirFilters; i++ ) {
	      
                DirFilterArray[i] = curDirPtr;
                curDirPtr += strlen( curDirPtr )+1;	    
            }

            break;

        case UNDELETE_MUTEX:

            //
            // The GUI has been started so we need to start worrying about synchronizing
            // access to the undelete bin. We only need to get the handle once; ignore
            // subsequent calls
            //
            if( !UndeleteMutex ) {

                ntStatus = ObReferenceObjectByHandle( (HANDLE) *(PULONG)InputBuffer, GENERIC_ALL,
                                                  NULL, KernelMode, &UndeleteMutex, NULL );
                if( !NT_SUCCESS( ntStatus )) {
                    
                    UndeleteMutex = NULL;
                }
                
                // 
                // We need to open this in the system process to that it won't go away
                //
                KeAttachProcess( SystemProcess );
                ntStatus = ObOpenObjectByPointer( UndeleteMutex, 0, NULL,
                                                  MUTANT_ALL_ACCESS, NULL, KernelMode, &mutexHandle );
                KeDetachProcess( );
            }
            break;
 
        default:

            //
            // Unknown control
            // 
            DbgPrint (("Undelete: unknown IRP_MJ_DEVICE_CONTROL\n"));
            IoStatus->Status = STATUS_INVALID_DEVICE_REQUEST;
            break;
        }

        retval = TRUE;

    } else {

        //
        // Its a call for a file system, so pass it through
        //
        hookExt = DeviceObject->DeviceExtension;

        if( hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoDeviceControl ) {
        
            retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoDeviceControl( 
                FileObject, Wait, InputBuffer, InputBufferLength, OutputBuffer, 
                OutputBufferLength, IoControlCode, IoStatus, 
                hookExt->FileSystem );
        }
    }

    return retval;
}


//----------------------------------------------------------------------
//
// UndeleteFastIoAcquireFile
//
//----------------------------------------------------------------------
VOID UndeleteFastIoAcquireFile( PFILE_OBJECT FileObject ) {
    PDEVICE_OBJECT      deviceObject;
    PHOOK_EXTENSION     hookExt;

    deviceObject = IoGetRelatedDeviceObject( FileObject );

    hookExt = deviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->AcquireFileForNtCreateSection ) {

        hookExt->FileSystem->DriverObject->FastIoDispatch->AcquireFileForNtCreateSection( 
            FileObject );
    }
}


//----------------------------------------------------------------------
//
// UndeleteFastIoReleaseFile
//
//----------------------------------------------------------------------
VOID UndeleteFastIoReleaseFile( PFILE_OBJECT FileObject ) {
    PDEVICE_OBJECT      deviceObject;
    PHOOK_EXTENSION     hookExt;

    deviceObject = IoGetRelatedDeviceObject( FileObject );

    hookExt = deviceObject->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->ReleaseFileForNtCreateSection ) {

        hookExt->FileSystem->DriverObject->FastIoDispatch->ReleaseFileForNtCreateSection( FileObject );

    }
}


//----------------------------------------------------------------------
//
// UndeleteFastIoDetachDevice
//
// We get this call when a device that we have hooked is being deleted.
// This happens when, for example, a floppy is formatted. We have
// to detach from it and delete our device. We should notify the GUI
// that the hook state has changed, but its not worth the trouble.
//
//----------------------------------------------------------------------
VOID UndeleteFastIoDetachDevice( PDEVICE_OBJECT SourceDevice, PDEVICE_OBJECT TargetDevice ) {
    PHOOK_EXTENSION     hookExt;
    ULONG               i;

    //
    // See if a device (like a floppy) is being removed out from under us. If so,
    // we have to detach from it before it disappears  
    //
    for( i = 2; i < 26; i++ ) {

        if( SourceDevice == LDriveDevices[i] ) {

            //
            // We've hooked it, so we must detach
            //
            hookExt = SourceDevice->DeviceExtension;

            DbgPrint(("Undelete: Detaching from drive: %c\n", 
                      hookExt->LogicalDrive ));

            IoDetachDevice( TargetDevice );
            IoDeleteDevice( SourceDevice );

            LDriveDevices[i] = NULL;

            return;
        }
    }

    //
    // Now we can pass the call through
    //
    hookExt = SourceDevice->DeviceExtension;

    if( hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoDetachDevice ) {

        hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoDetachDevice( 
            SourceDevice, TargetDevice );

    }
}

//----------------------------------------------------------------------
//           F I L E   U N D L E T E   R O U T I N E S
//----------------------------------------------------------------------


//----------------------------------------------------------------------
//
// UndeleteInitInfoFile
//
// Initialize a new INFO file.
//
//----------------------------------------------------------------------
VOID UndeleteInitInfoHeader( BOOLEAN IsAnsi, PINFO_HEADER InfoHeader )
{

    InfoHeader->Entries   = 1;  // anticipating the new entry
    InfoHeader->NextID    = 1;  
    InfoHeader->Reserved2 = 0;

    //
    // If the volume is marked as ANSI, init an ANSI bin
    //
    if( IsAnsi ) {

        InfoHeader->VersionSignature = ANSI_VERSION_SIG;
        InfoHeader->Version = ANSI_RECYCLE_TYPE;

    } else {

        InfoHeader->VersionSignature = UNICODE_VERSION_SIG;
        InfoHeader->Version = UNICODE_RECYCLE_TYPE;
    }
}


//----------------------------------------------------------------------
//
// UndeleteReadInfoFile
//
// Opens the INFO file in the directory passed in, and reads out
// the header to get the file ID that should be assigned to the next
// file to be moved to the recycle bin.
//
//----------------------------------------------------------------------
ULONG UndeleteReadInfoFile( BOOLEAN IsAnsi,
                            PWCHAR RecycleBinPath, 
                            PINFO_HEADER InfoHeader,
                            PINFO_LIST *EntryList )
{
    UNICODE_STRING          infoFileName;
    HANDLE                  infoFile;
    PWCHAR                  infoFileNameBuffer;
    OBJECT_ATTRIBUTES       objectAttributes;
    IO_STATUS_BLOCK  	    ioStatus;
    NTSTATUS                ntStatus;
    ULONG                   fileId;
    ULONG                   i;
    PINFO_LIST              nextEntry, prevEntry;

    //
    // Assume no entries in file
    //
    *EntryList = NULL;
    fileId = 0;

    //
    // Make the name of the INFO file
    //
    infoFileNameBuffer = ExAllocatePool( NonPagedPool, 1024 );
    swprintf( infoFileNameBuffer, L"%s\\INFO", RecycleBinPath );
    RtlInitUnicodeString( &infoFileName, infoFileNameBuffer );

    //
    // Open it.
    //
    InitializeObjectAttributes( &objectAttributes, &infoFileName,
                                OBJ_CASE_INSENSITIVE, NULL, NULL );

    ntStatus = ZwCreateFile( &infoFile, FILE_READ_DATA|SYNCHRONIZE,
                             &objectAttributes, &ioStatus, NULL, FILE_ATTRIBUTE_HIDDEN, 
                             FILE_SHARE_READ,
                             FILE_OPEN_IF, FILE_SYNCHRONOUS_IO_NONALERT, NULL, 0 );
    if( !NT_SUCCESS( ntStatus )) {

        //
        // For now, assume INFO file doesn't exist yet. Go ahead and initialize the header
        //
        UndeleteInitInfoHeader( IsAnsi, InfoHeader );        
        DbgPrint(("Undelete: no info file yet.\n"));
        return 0;
    } 

    //
    // Read the header out of the file
    //
    ntStatus = ZwReadFile( infoFile, NULL, NULL, NULL, &ioStatus, InfoHeader, 
                           sizeof( INFO_HEADER ), NULL, NULL );

    if( !NT_SUCCESS( ntStatus ) || !InfoHeader->Entries) {

        //
        // Hmmm...well, initialize a new header
        //
        UndeleteInitInfoHeader( IsAnsi, InfoHeader );
        ZwClose( infoFile );
        return 0;
    }

    //
    // Read each entry out of the file
    //
    prevEntry = NULL;
    for( i = 0; i < InfoHeader->Entries; i++ ) {

        nextEntry = ExAllocatePool( PagedPool, sizeof( INFO_LIST ));
        if( !nextEntry ) {

            //
            // Really bad if no more paged pool
            //
            DbgPrint(("Undelete: OUT OF PAGED POOL!\n"));
            break;
        }

        nextEntry->Next = NULL;

        //
        // Suck it in
        //
        if( InfoHeader->Version == UNICODE_RECYCLE_TYPE ) 
            ntStatus = ZwReadFile( infoFile, NULL, NULL, NULL, &ioStatus, &nextEntry->UniEntry,
                                   sizeof( UNIINFO_ENTRY), NULL, NULL );
        else 
            ntStatus = ZwReadFile( infoFile, NULL, NULL, NULL, &ioStatus, &nextEntry->AnsiEntry,
                                   sizeof( ANSIINFO_ENTRY), NULL, NULL );
        if( !NT_SUCCESS( ntStatus )) {

            //
            // File is shorter than we thought
            //
            DbgPrint(("Undelete: file is only %d entries instead of %d\n", i, InfoHeader->Entries ));
            InfoHeader->Entries = i;
            break;
        }
        
        //
        // Stick the entry on the end of the list
        //
        if( prevEntry ) {
   
            prevEntry->Next = nextEntry;
            prevEntry = nextEntry;

        } else {

            *EntryList = nextEntry;
            prevEntry = nextEntry;

        }

        //
        // Keep track of the highest iD
        //
        if( fileId <= nextEntry->UniEntry.FileID ) {

            fileId = nextEntry->UniEntry.FileID+1;
        }
    }

    //
    // Update the header to reflect the new addition.
    //
    InfoHeader->Entries++;
    InfoHeader->NextID = fileId+1;
    ZwClose( infoFile );
    return fileId;
}


//----------------------------------------------------------------------
//
// UndeleteWriteInfoFile
//
// Opens the INFO file in the directory passed in, and reads out
// the header to get the file ID that should be assigned to the next
// file to be moved to the recycle bin.
//
//----------------------------------------------------------------------
VOID UndeleteWriteInfoFile( PWCHAR RecycleBinPath, PINFO_HEADER InfoHeader,
                            PINFO_LIST EntryList, PINFO_LIST NewEntry )
{
    UNICODE_STRING          infoFileName;
    PWCHAR                  infoFileNameBuffer;
    OBJECT_ATTRIBUTES       objectAttributes;
    IO_STATUS_BLOCK  	    ioStatus;
    NTSTATUS                ntStatus;
    HANDLE                  infoFile;
    ULONG                   fileId;
    ULONG                   i;
    PINFO_LIST              nextEntry;

    //
    // Make the name of the INFO file
    //
    infoFileNameBuffer = ExAllocatePool( NonPagedPool, 1024 );
    swprintf( infoFileNameBuffer, L"%s\\INFO", RecycleBinPath );
    RtlInitUnicodeString( &infoFileName, infoFileNameBuffer );

    //
    // Open it.
    //
    InitializeObjectAttributes( &objectAttributes, &infoFileName,
                                OBJ_CASE_INSENSITIVE, NULL, NULL );

    ntStatus = ZwCreateFile( &infoFile, FILE_WRITE_DATA|SYNCHRONIZE,
                             &objectAttributes, &ioStatus, NULL, FILE_ATTRIBUTE_HIDDEN, 
                             FILE_SHARE_READ,
                             FILE_OPEN_IF, FILE_SYNCHRONOUS_IO_NONALERT, NULL, 0 );
    if( !NT_SUCCESS( ntStatus )) {

        //
        // Bad if we can't open or create it
        //
        DbgPrint(("Undelete: Could not open/create INFO file for writing:%x\n", ntStatus));
        return;
    }

    //
    // Write the header
    //
    ntStatus = ZwWriteFile( infoFile, NULL, NULL, NULL, &ioStatus, InfoHeader, 
                            sizeof( INFO_HEADER ), NULL, NULL );

    if( !NT_SUCCESS( ntStatus )) {

        //
        // Hmmm...that's bad
        //
        DbgPrint(("Undelete: could not update INFO file\n"));
        ZwClose( infoFile );
        return;
    }

    //
    // Write each entry to the file, freeing it when done
    //
    nextEntry = EntryList;
    for( i = 0; i < InfoHeader->Entries - 1; i++ ) {

        //
        // We ran out of entries early
        //
        if( !nextEntry ) break;

        if( InfoHeader->Version == UNICODE_RECYCLE_TYPE )
            ntStatus = ZwWriteFile( infoFile, NULL, NULL, NULL, &ioStatus, &nextEntry->UniEntry,
                                    sizeof( UNIINFO_ENTRY), NULL, NULL );
        else
            ntStatus = ZwWriteFile( infoFile, NULL, NULL, NULL, &ioStatus, &nextEntry->AnsiEntry,
                                    sizeof( ANSIINFO_ENTRY), NULL, NULL );
        if( !NT_SUCCESS( ntStatus )) {

            //
            // That's bad
            //
            DbgPrint(("Undelete: could not write to INFO file:%x\n", ntStatus ));
        }
        
        //
        // Free the entry
        //
        EntryList = nextEntry;
        nextEntry = nextEntry->Next;
        ExFreePool( EntryList );
    }

    //
    // Write out the new entry, if there is one (it would fit in the bin)
    // 
    if( NewEntry ) {

        if( InfoHeader->Version == UNICODE_RECYCLE_TYPE ) 
            ntStatus = ZwWriteFile( infoFile, NULL, NULL, NULL, &ioStatus, &NewEntry->UniEntry,
                                    sizeof( UNIINFO_ENTRY), NULL, NULL );
        else
            ntStatus = ZwWriteFile( infoFile, NULL, NULL, NULL, &ioStatus, &NewEntry->AnsiEntry,
                                    sizeof( ANSIINFO_ENTRY), NULL, NULL );
        if( !NT_SUCCESS( ntStatus )) {

            //
            // That's bad
            //
            DbgPrint(("Undelete: could not write new entry to INFO file:%x\n", ntStatus ));
        }
        ExFreePool( NewEntry );
    }

    ZwClose( infoFile );
    return;
}


//----------------------------------------------------------------------
//
// UndeleteRecycleCurrentSize
//
// Does a directory enumeration of the recycle bin in order to get
// its current size.
//
//----------------------------------------------------------------------
ULONGLONG UndeleteGetRecycleBinSize( UCHAR LogicalDrive, PWCHAR RecycleBinPath ) 
{
    UNICODE_STRING          recycleUnicodeString;
    OBJECT_ATTRIBUTES       objectAttributes;
    HANDLE                  dirHandle;
    NTSTATUS                ntStatus;
    IO_STATUS_BLOCK         ioStatus;
    BOOLEAN                 firstQuery = TRUE;
    ULONGLONG               binSize = 0;
    PFILE_DIRECTORY_INFORMATION dirInfo, curEntry;

    RtlInitUnicodeString( &recycleUnicodeString, RecycleBinPath );
    
    InitializeObjectAttributes( &objectAttributes, &recycleUnicodeString,
                                OBJ_CASE_INSENSITIVE, NULL, NULL );

    ntStatus = ZwCreateFile( &dirHandle, FILE_LIST_DIRECTORY,
                             &objectAttributes, &ioStatus, NULL, 0, FILE_SHARE_READ|FILE_SHARE_WRITE, 
                             FILE_OPEN, 
                             FILE_SYNCHRONOUS_IO_NONALERT|FILE_DIRECTORY_FILE, 
                             NULL, 0 );
    if( !NT_SUCCESS( ntStatus ) ) {

        DbgPrint(("Undelete: Could not open recycle bin: %x\n", ntStatus ));
        return (ULONGLONG) 0;
    }

    //
    // Allocate space for directory entry information
    //
    dirInfo = (PFILE_DIRECTORY_INFORMATION) ExAllocatePool( NonPagedPool, 4096 );

    while( 1 ) {

        //
        // Get the contents of the directory, adding up the size as we go
        //
        ntStatus = ZwQueryDirectoryFile( dirHandle, NULL, NULL, 0, &ioStatus, 
                                         dirInfo, 4096, FileDirectoryInformation,
                                         FALSE, NULL, firstQuery );
        if( !NT_SUCCESS( ntStatus )) {

            //
            // Got what we wanted
            //
            DbgPrint(("Undelete: bin dir enum terminated with %x\n", ntStatus ));

            ZwClose( dirHandle );
            ExFreePool( dirInfo );
            return binSize;
        }

        //
        // Add up the sizes
        //
        curEntry = dirInfo;
  
        while( 1 ) {

            binSize += (ULONGLONG) curEntry->AllocationSize.QuadPart;

            //
            // End of list is a 0 for the next offset
            //
            if( curEntry->NextEntryOffset == 0 ) break;

            curEntry = (PFILE_DIRECTORY_INFORMATION) ((PCHAR) curEntry + curEntry->NextEntryOffset);

        } 

        firstQuery = FALSE;
    }

    //
    // Never get here
    //
    return 0;
}

//----------------------------------------------------------------------
//
// UndeleteGetAnsiExtension
//
// Given a name of a file embedded in an INFO file entry, this routine
// seeks out the file extension.
//
//----------------------------------------------------------------------
PCHAR UndeleteGetAnsiExtension( PCHAR FileName )
{
    int           i = 0;
    int           length = 0;
    PCHAR        extPointer;

    //
    // Go to the end of the file name
    //
    while( FileName[i++] ) length++;
    extPointer = &FileName[length-1];

    //
    // Seek back looking for the extension
    //
    for(i = 0; i < length - 1; i++ ) {

        if( *extPointer == L'\\' || *extPointer == L'.' ) break;
        extPointer--;
    }

    //
    // Return the pointer
    //
    return extPointer;
}


//----------------------------------------------------------------------
//
// UndeleteGetUnicodeExtension
//
// Given a name of a file embedded in an INFO file entry, this routine
// seeks out the file extension.
//
//----------------------------------------------------------------------
PWCHAR UndeleteGetUnicodeExtension( PWCHAR FileName )
{
    int           i = 0;
    int           length = 0;
    PWCHAR        extPointer;

    //
    // Go to the end of the file name
    //
    while( FileName[i++] ) length++;
    extPointer = &FileName[length-1];

    //
    // Seek back looking for the extension
    //
    for(i = 0; i < length - 1; i++ ) {

        if( *extPointer == L'\\' || *extPointer == L'.' ) break;
        extPointer--;
    }

    //
    // Return the pointer
    //
    return extPointer;
}

//----------------------------------------------------------------------
//
// UndeleteGetAsciiExtension
//
// Given a name of a file, seeks out the ascii extension.
//
//----------------------------------------------------------------------
VOID UndeleteGetAsciiExtension( PWCHAR FileName, PCHAR AsciiName, PCHAR Extension )
{
    int           i = 0;
    int           length = 0;
    PWCHAR        extPointer;

    //
    // Determine the length of the name
    //
    while( FileName[i++] ) length++;

    //
    // First, make the ascii name
    //
    for( i = 0; i < length; i++ ) {
      
        AsciiName[i] = (CHAR) FileName[i];
    }
    AsciiName[i] = 0;

    //
    // Go to the end of the file name
    //
    while( FileName[i++] ) length++;
    extPointer = &FileName[length-1];

    //
    // Seek back looking for the extension
    //
    for(i = 0; i < length - 1; i++ ) {

        if( *extPointer == L'\\' || *extPointer == L'.' ) break;
        extPointer--;
    }

    i = 0;
    if( *extPointer == L'.' ) {

        while( *extPointer ) {

            Extension[i++] = *(PCHAR) extPointer;
            extPointer++;
        }
    }
    Extension[i] = 0;
}


//----------------------------------------------------------------------
//
// UndeleteDeleteSpillage
//
// This routine makes sure that the Recycle Bin hasn't overflowed
// its limits. If it has, then we have to really delete things out
// of it.
//
//----------------------------------------------------------------------
VOID UndeleteDeleteSpillage( PDEVICE_OBJECT FileSystem, UCHAR Volume, 
                             PWCHAR RecycleBinPath,
                             PPURGE_INFORMATION PurgeInfo,
                             PINFO_HEADER InfoHeader,
                             PINFO_LIST *EntryList, PINFO_LIST *NewEntry )
{
    WCHAR                   volumeBuffer[] = L"\\DosDevices\\X:\\";
    UNICODE_STRING          volumeBufferUnicodeString;
    HANDLE 		    volumeHandle;   
    OBJECT_ATTRIBUTES       objectAttributes;
    IO_STATUS_BLOCK  	    ioStatus;
    FILE_FS_SIZE_INFORMATION fsSizeInformation;
    ULONGLONG               recycleMaxSize;
    ULONGLONG		    recycleCurrentSize;
    NTSTATUS                ntStatus;
    PWCHAR                  fileNameBuffer;
    PWCHAR                  extPointer;
    ANSI_STRING             ansiExtension;
    UNICODE_STRING          extString;
    UNICODE_STRING          fileName;
    PINFO_LIST              deleteEntry, entryList;

    //
    // Open the volume
    //
    RtlInitUnicodeString (&volumeBufferUnicodeString,
                          volumeBuffer );    
    volumeBufferUnicodeString.Buffer[12] = Volume;

    InitializeObjectAttributes( &objectAttributes, &volumeBufferUnicodeString, 
                                OBJ_CASE_INSENSITIVE, NULL, NULL );

    ntStatus = ZwCreateFile( &volumeHandle, FILE_ANY_ACCESS, 
                             &objectAttributes, &ioStatus, NULL, 0, FILE_SHARE_READ|FILE_SHARE_WRITE, 
                             FILE_OPEN, 
                             FILE_SYNCHRONOUS_IO_NONALERT|FILE_OPEN_FOR_BACKUP_INTENT|FILE_DIRECTORY_FILE, 
                             NULL, 0 );
    if( !NT_SUCCESS( ntStatus ) ) {

        DbgPrint(("Undelete: Could not open drive %c: %x\n", Volume, ntStatus ));
        return;

    }
    
    //
    // Opened the volume. Now ask how much total space is on the drive
    //
    ntStatus = ZwQueryVolumeInformationFile( volumeHandle, &ioStatus, &fsSizeInformation,
                                             sizeof(fsSizeInformation), 
                                             FileFsSizeInformation );
    if( !NT_SUCCESS( ntStatus )) {

        DbgPrint(("Undelete: Could not get volume size: %x\n", ntStatus));
        return;
    }
    ZwClose( volumeHandle );

    //
    // Caclulate the max size of the bin
    //
    if( PurgeInfo->EnforceGlobalSetting ) {

        recycleMaxSize = ((ULONGLONG)fsSizeInformation.TotalAllocationUnits.QuadPart * 
                          (ULONGLONG)fsSizeInformation.BytesPerSector *
                          (ULONGLONG)fsSizeInformation.SectorsPerAllocationUnit *
                          (ULONGLONG)(PurgeInfo->ReservedForRecycle[GLOBAL_SETTING] & 0xFF)) / (ULONGLONG) 100;

    } else {

        recycleMaxSize = ((ULONGLONG)fsSizeInformation.TotalAllocationUnits.QuadPart * 
                          (ULONGLONG)fsSizeInformation.BytesPerSector *
                          (ULONGLONG)fsSizeInformation.SectorsPerAllocationUnit *
                          (ULONGLONG)(PurgeInfo->ReservedForRecycle[Volume - 'A'] & 0xFF)) 
            / (ULONGLONG) 100;
    }

    //
    // See how much space the Bin is currently taking up.
    //
    recycleCurrentSize = UndeleteGetRecycleBinSize( Volume, RecycleBinPath );
 
    DbgPrint(("%C Bin Max: %d current: %d\n", Volume, recycleMaxSize, recycleCurrentSize ));

    //
    // Allocate space for a filename
    //
    fileNameBuffer = (PWCHAR) ExAllocatePool( NonPagedPool, 2048 );

    //
    // Okay, empty files out of the bin until we drop below the size limit
    //
    while( recycleCurrentSize > recycleMaxSize ) {
 
        DbgPrint(("Undelete: trimming a file (%d present)\n", InfoHeader->Entries ));

        //
        // If there are old entries, delete one
        //
        if( InfoHeader->Entries - 1 ) {

            deleteEntry = *EntryList;
            *EntryList = (*EntryList)->Next;

        } else {

            deleteEntry = *NewEntry;
            *NewEntry = NULL;
        }

        //
        // Make the filename. Here it depends on whether we're dealing with 
        // a unicode or ansi recycle bin. Ansi recycle bins are present if
        // the user left things in the bin in Win95 before dual booting into NT
        //
        if( InfoHeader->Version == UNICODE_RECYCLE_TYPE ) {

            extPointer = UndeleteGetUnicodeExtension( deleteEntry->UniEntry.UnicodeName );
            swprintf( fileNameBuffer, L"%s\\D%C%d%s",
                      RecycleBinPath, Volume, deleteEntry->UniEntry.FileID,
                      *extPointer == L'.' ? extPointer : L"" );
            RtlInitUnicodeString( &fileName, fileNameBuffer );

        } else {

            ansiExtension.Buffer = UndeleteGetAnsiExtension( deleteEntry->AnsiEntry.AsciiName );
            ansiExtension.MaximumLength = 2*strlen(ansiExtension.Buffer)+2;
            ansiExtension.Length = ansiExtension.MaximumLength;

            RtlAnsiStringToUnicodeString( &extString, &ansiExtension, TRUE );
            swprintf( fileNameBuffer, L"%s\\D%C%d%s",
                      RecycleBinPath, Volume, deleteEntry->AnsiEntry.FileID,
                      *extString.Buffer == L'.' ? extString.Buffer : L"" );
            RtlInitUnicodeString( &fileName, fileNameBuffer );
        }

        //
        // Delete the file, getting the size subtracted
        //
        recycleCurrentSize -= UndeleteDeleteFile( FileSystem, &fileName );

        //
        // Free the entry
        //
        ExFreePool( deleteEntry );

        //
        // Anything more to delete?
        //
        if( !(--InfoHeader->Entries) ) {

            break;
        }
    }

    //
    // Free the buffer
    //
    ExFreePool( fileNameBuffer );
}


//----------------------------------------------------------------------
//
// UndeleteFile
//
// This is the work routine for moving stuff into the recycle bin.
//
//----------------------------------------------------------------------
VOID UndeleteFile( PUNDELETE_COMPLETE_CONTEXT CompleteContext )
{
    PHOOK_EXTENSION         hookExt;
    NTSTATUS                ntStatus;
    OBJECT_ATTRIBUTES       objectAttributes;
    ULONG                   index;
    HANDLE                  hKey;
    ULONG                   length;
    WCHAR                   HKUName[] = L"\\Registry\\USER";
    UNICODE_STRING          sidString;
    ULONG                   requiredLength;
    PCHAR                   sidStringBuffer[512];
    UNICODE_STRING          HKUUnicodeString;
    UNICODE_STRING          binName;
    PTOKEN_USER             tokenInfoBuffer;
    HANDLE                  tokenHandle;
    PVOID                   token;
    PWCHAR                  binNameBuffer;
    HANDLE                  dirHandle;
    IO_STATUS_BLOCK  	    ioStatus;    
    PINFO_HEADER            infoHeader;
    PINFO_LIST              infoList;
    PINFO_LIST              newEntry;
    PINFO_LIST              nextEntry;
    PWCHAR                  newNameBuffer;
    LONG                    i;
    PWCHAR                  extPointer;

    //
    // Get the hook extension
    //
    hookExt = CompleteContext->DeviceObject->DeviceExtension;

    //
    // Allocate the name
    //
    binNameBuffer = ExAllocatePool( NonPagedPool, 1024 );

    //
    // Construct the path to the recycle directory this file is going into.
    // We have to get user's SID if this is a delete on an NTFS volume
    //
    if( hookExt->IsNTFS ) {

        //
        // Make sure the top-level recycle directory exists
        //
        swprintf( binNameBuffer, L"\\??\\%C:\\NTRECYCLER", hookExt->LogicalDrive );
        RtlInitUnicodeString( &binName, binNameBuffer );
        InitializeObjectAttributes( &objectAttributes, &binName, 
                                    OBJ_CASE_INSENSITIVE, NULL, NULL );
        ntStatus = ZwCreateFile( &dirHandle, GENERIC_WRITE,
                                 &objectAttributes, &ioStatus, NULL, FILE_ATTRIBUTE_HIDDEN, 
                                 FILE_SHARE_READ|FILE_SHARE_WRITE,
                                 FILE_CREATE, FILE_DIRECTORY_FILE, NULL, 0 );
        if( !NT_SUCCESS( ntStatus )) {

            if( ntStatus != STATUS_OBJECT_NAME_COLLISION ) {

                //
                // Error that prevents us from backing up files
                //
                DbgPrint(("Undelete: Could not make RECYCLER subdirectory for volume %C\n", hookExt->LogicalDrive ));
                ExFreePool( binNameBuffer );
                return;
            } 

        } else {

            ZwClose( dirHandle );
        }

        //
        // Get a handle for the token
        //
        ntStatus = ObOpenObjectByPointer( CompleteContext->Token, 0, NULL,
                                          TOKEN_QUERY, NULL, KernelMode, &tokenHandle );
        ObDereferenceObject( CompleteContext->Token );
        if( !NT_SUCCESS( ntStatus )) {

            DbgPrint(("Undelete: Could not open process token: %x\n", ntStatus ));
            ExFreePool( binNameBuffer );
            return;
        }        
        
        //
        // Pull out the SID
        //
        ntStatus = NtQueryInformationToken( tokenHandle, TokenUser, NULL, 0, &requiredLength );
        if(  ntStatus != STATUS_BUFFER_TOO_SMALL ) {

            DbgPrint(("Undelete: Error getting token information: %x\n", ntStatus));
            ExFreePool( binNameBuffer );
            ZwClose( tokenHandle );
            return;
        }
        tokenInfoBuffer = (PTOKEN_USER) ExAllocatePool( NonPagedPool, requiredLength );
        ntStatus = NtQueryInformationToken( tokenHandle, TokenUser, tokenInfoBuffer, requiredLength, &requiredLength );
        if( !NT_SUCCESS( ntStatus )) {

            DbgPrint(("Undelete: Error getting token information: %x\n", ntStatus));
            ExFreePool( tokenInfoBuffer );
            ExFreePool( binNameBuffer );
            ZwClose( tokenHandle );
            return;
        }
        ZwClose( tokenHandle );

        //
        // Got it, now convert to text representation
        //
        memset( sidStringBuffer, 0, sizeof(sidStringBuffer ));
        sidString.Buffer = (PWCHAR) sidStringBuffer;
        sidString.MaximumLength = sizeof(sidStringBuffer);
        ntStatus = RtlConvertSidToUnicodeString( &sidString, tokenInfoBuffer->User.Sid, FALSE );
        ExFreePool( tokenInfoBuffer );
        if( !NT_SUCCESS( ntStatus )) {

            DbgPrint(("Undelete: Unable to convert SID to text: %x\n", ntStatus ));
            ExFreePool( binNameBuffer );
            return;
        }            

        //
        // We can make the name now
        //
        swprintf( binNameBuffer, L"\\??\\%C:\\NTRECYCLER\\%s", hookExt->LogicalDrive, 
                  sidString.Buffer );

        //
        // Make sure the SID sub-directory exists.
        // FIX: We need to capture the subject security context here 
        //
        RtlInitUnicodeString( &binName, binNameBuffer );
        InitializeObjectAttributes( &objectAttributes, &binName, 
                                    OBJ_CASE_INSENSITIVE, NULL, NULL );
        ntStatus = ZwCreateFile( &dirHandle, GENERIC_WRITE,
                                 &objectAttributes, &ioStatus, NULL, FILE_ATTRIBUTE_HIDDEN, 
                                 FILE_SHARE_READ|FILE_SHARE_WRITE,
                                 FILE_CREATE, FILE_DIRECTORY_FILE, NULL, 0 );
        if( !NT_SUCCESS( ntStatus )) {

            if( ntStatus != STATUS_OBJECT_NAME_COLLISION ) {

                //
                // Error that prevents us from backing up files
                //
                DbgPrint(("Undelete: Could not make SID subdirectory for volume %C\n", hookExt->LogicalDrive ));
                ExFreePool( binNameBuffer );
                return;
            }
        } else {

            ZwClose( dirHandle );
        }

    } else {
   
        //
        // Make sure the top-level recycle directory exists (its also the full path of the bin)
        //
        swprintf( binNameBuffer, L"\\??\\%C:\\NTRECYCLED", hookExt->LogicalDrive );
        RtlInitUnicodeString( &binName, binNameBuffer );
        InitializeObjectAttributes( &objectAttributes, &binName, 
                                    OBJ_CASE_INSENSITIVE, NULL, NULL );
        ntStatus = ZwCreateFile( &dirHandle, GENERIC_WRITE,
                                 &objectAttributes, &ioStatus, NULL, FILE_ATTRIBUTE_HIDDEN, 
                                 FILE_SHARE_READ|FILE_SHARE_WRITE,
                                 FILE_CREATE, FILE_DIRECTORY_FILE, NULL, 0 );
        if( !NT_SUCCESS( ntStatus )) {

            if( ntStatus != STATUS_OBJECT_NAME_COLLISION ) {

                //
                // Error that prevents us from backing up files
                //
                DbgPrint(("Undelete: Could not make RECYCLED subdirectory for volume %C\n", hookExt->LogicalDrive ));
                ExFreePool( binNameBuffer );
                return;
            }
        } else {

            ZwClose( dirHandle );
        }
    }

    DbgPrint(("Moving this guy to the recycle bin\n"));

    //
    // Create the new entry and get the INFO header so that we can see
    //  what ID this file is getting
    //
    newEntry = ExAllocatePool( PagedPool, sizeof( INFO_LIST ));
    memset( newEntry, 0, sizeof( INFO_LIST ));
    infoHeader = (PINFO_HEADER) ExAllocatePool( NonPagedPool, sizeof( INFO_HEADER ));
    newEntry->UniEntry.FileID = UndeleteReadInfoFile( hookExt->IsAnsi, binNameBuffer, infoHeader, &infoList );
    newEntry->UniEntry.Drive  = hookExt->LogicalDrive - 'A';
    KeQuerySystemTime( &newEntry->UniEntry.TimeDeleted );
    DbgPrint(("   file id is: %d\n", newEntry->UniEntry.FileID ));

    //
    // Remember if the bin is ansi
    //
    hookExt->IsAnsi = (infoHeader->Version == ANSI_RECYCLE_TYPE);

    //
    // If this is simply a check to see what type of recycle bin we're dealing with
    // (which happens during initialization) return now.
    //
    if( !CompleteContext->Irp ) {

        DbgPrint(("%C: checked recycle bin and its type is %s\n",
                  hookExt->LogicalDrive, hookExt->IsAnsi ? "ANSI" : "UNICODE" ));
        while( infoList ) {

            nextEntry = infoList->Next;
            ExFreePool( infoList );
            infoList = nextEntry;
        }
        ExFreePool( binNameBuffer );
        ExFreePool( infoHeader );
        ExFreePool( newEntry );
        return;
    }    

    //
    // Construct the new name for the file - we have to preserve the extension if the file
    // has one
    //
    newNameBuffer = ExAllocatePool( PagedPool, 1024 );
    extPointer = &CompleteContext->FileName.Buffer[CompleteContext->FileName.Length/2-1];
    for(i = 0; i < CompleteContext->FileName.Length/2; i++ ) {

        if( *extPointer == L'\\' || *extPointer == L'.' ) break;
        extPointer--;
    }
    swprintf( newNameBuffer, L"%s\\D%C%d%s", binNameBuffer, hookExt->LogicalDrive,
              newEntry->UniEntry.FileID, *extPointer == L'.'  ? extPointer : L"" );

    //
    // Stick the old in the new entry (except for the leading \??\) (note pointer arithmetic)
    // Only do Unicode for the unicode version
    //
    extPointer = CompleteContext->FileName.Buffer + 4;

    if( infoHeader->Version == UNICODE_RECYCLE_TYPE ) {

        wcscpy( newEntry->UniEntry.UnicodeName, extPointer );
    }

    i = 0;
    while( *extPointer ) {

        newEntry->UniEntry.AsciiName[i++] = (UCHAR) *extPointer;
        extPointer++;
    }

    //
    // Do the rename
    //
    UndeleteRenameFile( &CompleteContext->FileName, newNameBuffer );

    //
    // Delete stuff out of the Recycle Bin if necessary in order to 
    // limit its size to what the user specified.
    //
    UndeleteDeleteSpillage( hookExt->FileSystem, hookExt->LogicalDrive, 
                            binNameBuffer, 
                            &CompleteContext->PurgeInfo, infoHeader,
                            &infoList, &newEntry );
 

    //
    // Dump out the updated file. The entry list is freed in this routine.
    //
    UndeleteWriteInfoFile( binNameBuffer, infoHeader, infoList, newEntry );

    //	
    // Clean up
    //
    ExFreePool( newNameBuffer );
    ExFreePool( binNameBuffer );
    ExFreePool( infoHeader );
}



//----------------------------------------------------------------------
//
// UndeleteCheckFileForUndelete
//
// See if this file is marked for delete. If so, get its name 
// and see if this is a file that should be protected.
//
//----------------------------------------------------------------------
BOOLEAN UndeleteCheckFileForUndelete( PDEVICE_OBJECT DeviceObject, PIRP Irp,
                                      PUNDELETE_COMPLETE_CONTEXT CompleteContext )
{
    PIO_STACK_LOCATION      currentIrpStack = IoGetCurrentIrpStackLocation(Irp);    
    PHOOK_EXTENSION         hookExt;
    NTSTATUS                ntStatus;
    CHAR                    asciiExtension[256];
    RTL_QUERY_REGISTRY_TABLE QueryTable[2];
    PURGE_INFORMATION	    purgeInfo;
    ULONG                   i;
    PCHAR                   asciiName;
    PCHAR                   asciiPtr;
    PCHAR                   fileNameBuffer;
    UNICODE_STRING          fileNameUnicodeString;
    UNICODE_STRING          upcaseFileName;

    //
    // Get the hook extension
    //
    hookExt = DeviceObject->DeviceExtension;

    //
    // First, make sure its not a directory
    //
    if( UndeleteIsDirectory( hookExt->FileSystem, currentIrpStack->FileObject ) ) {
    
        return FALSE;
    }

    //
    // Read the recycle bin configuration from the registry
    //
    *(ULONG *) &purgeInfo = sizeof(PURGE_INFORMATION);
    memset( QueryTable, 0, sizeof(RTL_QUERY_REGISTRY_TABLE) * 2);
    QueryTable[0].QueryRoutine = NULL;
    QueryTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT;
    QueryTable[0].EntryContext = &purgeInfo;
    QueryTable[0].Name = L"PurgeInfo";
    QueryTable[0].DefaultType = REG_BINARY;
    QueryTable[0].DefaultData = &purgeInfo;
    QueryTable[0].DefaultLength = sizeof(PURGE_INFORMATION);
    ntStatus = RtlQueryRegistryValues( RTL_REGISTRY_ABSOLUTE, 
                                       L"\\Registry\\Machine\\SOFTWARE\\Systems Internals\\Undelete", 
                                       QueryTable, NULL, NULL );
    
    //
    // If we can't read the Purge info, it means that everything is undeleted
    //
    if( !NT_SUCCESS( ntStatus )) {

        purgeInfo.EnforceGlobalSetting = 1;
        purgeInfo.RecycleDisableMask = 0;
        purgeInfo.ReservedForRecycle[GLOBAL_SETTING] = 100; 
    }

    //
    // Are we doing undeletes for this volume?
    //
    if( (purgeInfo.EnforceGlobalSetting && (purgeInfo.RecycleDisableMask & 1 << GLOBAL_SETTING)) ||
        (!purgeInfo.EnforceGlobalSetting && (purgeInfo.RecycleDisableMask & 1 << hookExt->LogicalDrive - 'A' ))) {

        DbgPrint(("Undelete: volume %C is not being protected\n", hookExt->LogicalDrive ));
        return FALSE;
    }

    //
    // Get the name of the file. We have to make sure this isn't a recycle bin
    // file getting deleted (we don't want to undelete those!)
    //
    fileNameBuffer = ExAllocatePool( PagedPool, 1024 );
    if( !UndeleteGetFileName( hookExt->IsNTFS, hookExt->FileSystem, currentIrpStack->FileObject,
                              fileNameBuffer, 1024)) {

        //
        // Couldn't get the name for some reason
        //
        ExFreePool( fileNameBuffer );
        return FALSE;
    }

    //
    // Is it in the Recycle Bin already. We are conservative about this as we 
    // only check the first component of the name. If its "recycler" or "recycled"
    // we go ahead and assume that its the user (or Explorer) emptying something from the bin.
    // When we allocate the buffer, leave room for the '\??\X:' (6 wide characters equals 12 bytes)
    // and the terminating null.
    //
    fileNameUnicodeString.Length = (USHORT) ((PFILE_NAME_INFORMATION) fileNameBuffer)->FileNameLength + 12;
    fileNameUnicodeString.MaximumLength = fileNameUnicodeString.Length;
    fileNameUnicodeString.Buffer = ExAllocatePool( PagedPool, fileNameUnicodeString.Length+2+12);
    swprintf( fileNameUnicodeString.Buffer, L"\\??\\%C:", hookExt->LogicalDrive );
    RtlCopyMemory( &fileNameUnicodeString.Buffer[6], ((PFILE_NAME_INFORMATION) fileNameBuffer)->FileName, 
                   fileNameUnicodeString.Length);
    fileNameUnicodeString.Buffer[fileNameUnicodeString.Length/2] = 0;

    upcaseFileName.Length = 0;
    upcaseFileName.MaximumLength = 256;
    upcaseFileName.Buffer = ExAllocatePool( NonPagedPool, 256 );
    RtlUpcaseUnicodeString( &upcaseFileName, &fileNameUnicodeString, FALSE );

    //
    // Do the comparison - remember to ignore the leading '\??\X:' (6 wide characters)
    //
    if( (hookExt->IsNTFS && !wcsncmp( &upcaseFileName.Buffer[6], L"\\NTRECYCLER", 9 )) ||
        ( !wcsncmp( &upcaseFileName.Buffer[6], L"\\NTRECYCLED", 9 ))) {

        //
        // Its in the recycler - don't do anything with it.
        //
        DbgPrint(("Undelete: its trash being emptied\n"));
        ExFreePool( upcaseFileName.Buffer );
        ExFreePool( fileNameBuffer );
        ExFreePool( fileNameUnicodeString.Buffer );
        return FALSE;
    }

    //
    // Convert the name to ascii
    //
    asciiName = ExAllocatePool( NonPagedPool, upcaseFileName.Length );
    UndeleteGetAsciiExtension( fileNameUnicodeString.Buffer, asciiName, asciiExtension );
    
    //
    // See if path is being filtered (ignore the leading '\??\' (
    //
    asciiPtr = &asciiName[4];
    for( i = 0; i < NumDirFilters; i++ ) {

        if( strlen( asciiPtr ) >= strlen( DirFilterArray[i]) ) {

            if( !_strnicmp( asciiPtr, DirFilterArray[i], strlen( DirFilterArray[i]) ) &&
                ( asciiPtr[ strlen(DirFilterArray[i])] == 0 || 
                  asciiPtr[ strlen(DirFilterArray[i])] == '\\')) {

                DbgPrint(("This directory (%s) is being filtered\n", DirFilterArray[i] ));
                ExFreePool( fileNameBuffer );
                ExFreePool( upcaseFileName.Buffer );          
                ExFreePool( asciiName );	    
                return FALSE;
            }
        }
    }

    //
    // Finally, see if its extension is being filtered. I used to think that extensions
    // could be a maximum of 3 characters in length, but I was wrong!
    //
    if( asciiExtension[0] ) {

        for( i = 0; i < NumFilters; i++ ) {

            if( !_stricmp( asciiExtension, FilterList[i].filter )) {

                //
                // We're not undeleting these files
                //
                DbgPrint(("This file (%s) extension is being filtered\n",
                          asciiExtension ));
                ExFreePool( fileNameBuffer );
                ExFreePool( upcaseFileName.Buffer );          
                ExFreePool( asciiName );
                return FALSE;
            }
        }  
    }

    //
    // Free buffers
    //
    ExFreePool( asciiName );
    ExFreePool( fileNameBuffer );
    ExFreePool( upcaseFileName.Buffer );

    //
    // Set up the completion context, where the rest of the undelete will take place
    //
    CompleteContext->FileName  = fileNameUnicodeString;
    CompleteContext->PurgeInfo = purgeInfo;
    return TRUE;
}

//----------------------------------------------------------------------
// 
// Undeleteworkitem
//
// Called by the system, this is initiated in the completion callback.
// We use it to ensure we''re running at passive level when we do
// an undelete
//
//----------------------------------------------------------------------
VOID UndeleteWorkItem( PVOID Context )
{
    PHOOK_EXTENSION     hookExt;
    BOOLEAN             mutexGrabbed = FALSE;

    //
    // Grab our mutex
    //
    if( UndeleteMutex ) {

        KeWaitForSingleObject( UndeleteMutex, Executive, KernelMode,
                               FALSE, NULL );
        mutexGrabbed = TRUE;
        DbgPrint(("Got mutex\n"));
    }

    //
    // Call the undelete routine
    //
    UndeleteFile( (PUNDELETE_COMPLETE_CONTEXT) Context );

    //
    // Release Mutex
    //
    if( mutexGrabbed ) {

        DbgPrint(("Releasing mutex\n"));        
        KeReleaseMutex( UndeleteMutex, FALSE );
    }

    //
    // Complete the IRP
    //
    ((PUNDELETE_COMPLETE_CONTEXT) Context)->Irp->IoStatus.Status = STATUS_SUCCESS;

    IoCompleteRequest( ((PUNDELETE_COMPLETE_CONTEXT) Context)->Irp, 
                       IO_NO_INCREMENT );

    //
    // Free memory
    //
    ExFreePool( ((PUNDELETE_COMPLETE_CONTEXT) Context)->FileName.Buffer );
    ExFreePool( Context );
}

//----------------------------------------------------------------------
//     D I S P A T C H   A N D   H O O K   E N T R Y   P O I N T S
//----------------------------------------------------------------------


//----------------------------------------------------------------------
// 
// UndeleteHookDone
//
// Gets control on the way back from a delete operation. At this point
// the file cannot be opened by anything else, so we can move it into
// the recycle bin.
//
//----------------------------------------------------------------------
NTSTATUS UndeleteHookDone( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp,
                           IN PVOID Context )
{
    if( NT_SUCCESS(Irp->IoStatus.Status)) {

        ExInitializeWorkItem( &((PUNDELETE_COMPLETE_CONTEXT) Context)->WorkItem, 
                              UndeleteWorkItem, Context );
        ExQueueWorkItem( &((PUNDELETE_COMPLETE_CONTEXT) Context)->WorkItem, 
                         CriticalWorkQueue );

        //
        // We have to complete the IRP later
        //
        return STATUS_MORE_PROCESSING_REQUIRED;

    } else {

        //
        // Now we have to mark Irp as pending if necessary
        //
        if( Irp->PendingReturned ) {

            IoMarkIrpPending( Irp );
        }

        // 
        // Free the completion context
        //
        ExFreePool( ((PUNDELETE_COMPLETE_CONTEXT) Context)->FileName.Buffer );
        ExFreePool( Context );
        return Irp->IoStatus.Status;
    }
}


//----------------------------------------------------------------------
//
// UndeleteHookRoutine
//
// This routine is the main hook routine where we figure out what
// calls are being sent to the file system.
//
//----------------------------------------------------------------------
NTSTATUS UndeleteHookRoutine( PDEVICE_OBJECT HookDevice, IN PIRP Irp )
{
    PIO_STACK_LOCATION  currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
    PIO_STACK_LOCATION  nextIrpStack    = IoGetNextIrpStackLocation(Irp);
    PFILE_OBJECT        fileObject;
    PHOOK_EXTENSION     hookExt;
    PUNDELETE_COMPLETE_CONTEXT completeContext;

    //
    // Extract the file object from the IRP
    //
    fileObject    = currentIrpStack->FileObject;

    //
    // Point at the device extension, which contains information on which
    // file system this IRP is headed for
    //
    hookExt = HookDevice->DeviceExtension;

    //
    // Copy parameters down to next level in the stack for the driver below us
    //
    *nextIrpStack = *currentIrpStack;

    //
    // Determine what function we're dealing with
    //
    switch( currentIrpStack->MajorFunction ) {

    case IRP_MJ_CLEANUP:

        //
        // If the file will be deleted, see if we have to undelete it
        //
        completeContext = ExAllocatePool( NonPagedPool, sizeof( UNDELETE_COMPLETE_CONTEXT));
        if( fileObject->DeletePending && (SystemProcess != PsGetCurrentProcess()) &&
            UndeleteCheckFileForUndelete( HookDevice, Irp, completeContext )) {

            //
            // Undelete the file
            //
            UndeleteSetDispositionFile( hookExt->FileSystem, currentIrpStack->FileObject,
                                        FALSE );

            //
            // Finish setting up the completion context
            //
            completeContext->DeviceObject = HookDevice;
            completeContext->Irp = Irp;
            completeContext->Token = PsReferencePrimaryToken(PsGetCurrentProcess());
            if( !completeContext->Token ) {

                //
                // Couldn't reference the token(?)
                //
                DbgPrint(("Undelete: Could not reference the process token.\n"));
                ExFreePool( completeContext );
                
            } else {

                //
                // Set the completion routine where we finish the job
                //
                DbgPrint(("Undelete: going to move this to the recycle bin\n"));

                IoSetCompletionRoutine( Irp, UndeleteHookDone, (PVOID) completeContext, TRUE, TRUE, TRUE );  
                
                IoMarkIrpPending( Irp );
                IoCallDriver( hookExt->FileSystem, Irp );

                //
                // Return this because we're going to stall the IRP in the 
                // completion routine
                //
                return STATUS_PENDING;
            }

        } else {
 
            ExFreePool( completeContext );
        }
        break;

    default:

        //
        // Register a never-called completion routine just so that the IRP stack
        // doesn't have a bogus completion routine in it.
        //
        IoSetCompletionRoutine( Irp, UndeleteHookDone, 0, FALSE, FALSE, FALSE );
        break;
    }

    //
    // Return the results of the call to the caller
    //
    return IoCallDriver( hookExt->FileSystem, Irp );
}



//----------------------------------------------------------------------
//
// UndeleteDeviceRoutine
//
// In this routine we handle requests to our own device. The only 
// requests we care about handling explicitely are IOCTL commands that
// we will get from the GUI. We also expect to get Create and Close 
// commands when the GUI opens and closes communications with us.
//
//----------------------------------------------------------------------
NTSTATUS UndeleteDeviceRoutine( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp )
{
    PIO_STACK_LOCATION  irpStack;
    PVOID               inputBuffer;
    PVOID               outputBuffer;
    ULONG               inputBufferLength;
    ULONG               outputBufferLength;
    ULONG               ioControlCode;

    //
    // Go ahead and set the request up as successful
    //
    Irp->IoStatus.Status      = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    //
    // Get a pointer to the current location in the Irp. This is where
    // the function codes and parameters are located.
    //
    irpStack = IoGetCurrentIrpStackLocation (Irp);

    //
    // Get the pointer to the input/output buffer and its length
    //
    inputBuffer     = Irp->AssociatedIrp.SystemBuffer;
    inputBufferLength   = irpStack->Parameters.DeviceIoControl.InputBufferLength;
    outputBuffer    = Irp->AssociatedIrp.SystemBuffer;
    outputBufferLength  = irpStack->Parameters.DeviceIoControl.OutputBufferLength;
    ioControlCode   = irpStack->Parameters.DeviceIoControl.IoControlCode;

    switch (irpStack->MajorFunction) {
    case IRP_MJ_CREATE:

        DbgPrint(("Undelete: IRP_MJ_CREATE\n"));
        break;

    case IRP_MJ_CLOSE:

        DbgPrint(("Undelete: IRP_MJ_CLOSE\n"));
        break;
    }

    //
    // Complete the IRP
    //
    IoCompleteRequest( Irp, IO_NO_INCREMENT );
    return STATUS_SUCCESS;   
}


//----------------------------------------------------------------------
//
// UndeleteDispatch
//
// Based on which device the Irp is destined for we call either the
// filesystem filter function, or our own device handling routine.
//
//----------------------------------------------------------------------
NTSTATUS UndeleteDispatch( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp )
{
    //
    // Determine if its a request from the GUI to us, or one that is
    // directed at a file system driver that we've hooked
    //
    if ( GUIDevice != DeviceObject ) {

        return( UndeleteHookRoutine( DeviceObject, Irp ));

    } else {

        return( UndeleteDeviceRoutine( DeviceObject, Irp ));

    }
}

//----------------------------------------------------------------------
//   D R I V E R - E N T R Y   A N D   H O O K   R O U T I N E S
//----------------------------------------------------------------------

//----------------------------------------------------------------------
//
// HookDrive
//
// Hook the drive specified by determining which device object to 
// attach to. The algorithm used here is similar to the one used
// internally by NT to determine which device object a file system request
// is directed at.
//
//----------------------------------------------------------------------
BOOLEAN HookDrive( IN USHORT Drive, IN PDRIVER_OBJECT DriverObject, 
                   IN HANDLE VolumeHandle )
{
    PDEVICE_OBJECT      fileSysDevice;
    PDEVICE_OBJECT      hookDevice;
    NTSTATUS            ntStatus;
    PFILE_OBJECT        fileObject;
    IO_STATUS_BLOCK  	ioStatus;    
    PHOOK_EXTENSION     hookExtension;
    UCHAR               fsAttribInformation[256];
    UNDELETE_COMPLETE_CONTEXT completeContext;

    //
    // Got the file handle, so now look-up the file-object it refers to
    //
    ntStatus = ObReferenceObjectByHandle( VolumeHandle, FILE_READ_DATA, 
                                          NULL, KernelMode, &fileObject, NULL );
    if( !NT_SUCCESS( ntStatus )) {

        DbgPrint(("Undelete: Could not get fileobject from handle: %c\n", 'A'+Drive ));
        return FALSE;
    }

    //  
    // Next, find out what device is associated with the file object by getting its related
    // device object
    //
    fileSysDevice = IoGetRelatedDeviceObject( fileObject );

    if ( ! fileSysDevice ) {

        DbgPrint(("Undelete: Could not get related device object: %c\n", 'A'+Drive ));
        ObDereferenceObject( fileObject );
        return FALSE;
    }

    //
    // The file system's device hasn't been hooked already, so make a hooking device
    //  object that will be attached to it.
    //
    ntStatus = IoCreateDevice( DriverObject,
                               sizeof(HOOK_EXTENSION),
                               NULL,
                               fileSysDevice->DeviceType,
                               0,
                               FALSE,
                               &hookDevice );

    if ( !NT_SUCCESS(ntStatus) ) {

        DbgPrint(("Undelete: failed to create associated device: %c\n", 'A'+Drive ));   
        ObDereferenceObject( fileObject );
        return FALSE;
    }

    //
    // Clear the device's init flag as per NT DDK KB article on creating device 
    // objects from a dispatch routine
    //
    hookDevice->Flags &= ~DO_DEVICE_INITIALIZING;

    //
    // Setup the device extensions. The drive letter and file system object are stored
    // in the extension.
    //
    hookExtension = hookDevice->DeviceExtension;
    hookExtension->LogicalDrive = 'A'+Drive;
    hookExtension->FileSystem   = fileSysDevice;
    hookExtension->IsNTFS       = FALSE;
    hookExtension->IsAnsi       = FALSE;

    //
    // Determine which file system this is. If this fails, assume its NOT NTFS
    //
    ntStatus = ZwQueryVolumeInformationFile( VolumeHandle, &ioStatus, fsAttribInformation,
                                             sizeof(fsAttribInformation), FileFsAttributeInformation );
    if( !NT_SUCCESS( ntStatus )) {

        DbgPrint(("Undelete: Could not get volume attributes: %x\n", ntStatus));

    } else if( RtlCompareMemory( ((PFILE_FS_ATTRIBUTE_INFORMATION)fsAttribInformation)->FileSystemName, L"NTFS", 8 ) == 8) {

        hookExtension->IsNTFS = TRUE;
        DbgPrint(("  => Its an NTFS volume\n"));
    } 

    //
    // Finally, attach to the device. The second we're successfully attached, we may 
    // start receiving IRPs targetted at the device we've hooked.
    //
    ntStatus = IoAttachDeviceByPointer( hookDevice, fileSysDevice );

    if ( !NT_SUCCESS(ntStatus) )  {

        //
        // Couldn't attach for some reason
        //
        DbgPrint(("Undelete: Connect with Filesystem failed: %c (%x) =>%x\n", 
                  'A'+Drive, fileSysDevice, ntStatus ));

        //
        // Derefence the object and get out
        //
        ObDereferenceObject( fileObject );
        return FALSE;

    } 

    //
    // Note which device this is
    //
    LDriveDevices[Drive] = hookDevice;

    //
    // Dereference the object
    //
    ObDereferenceObject( fileObject );

    //
    // Now see what type of bin it is
    //
    completeContext.DeviceObject = hookDevice;
    completeContext.Irp = NULL;  // this indicates its a check
    completeContext.Token = PsReferencePrimaryToken(PsGetCurrentProcess());
    UndeleteFile( &completeContext );
    return TRUE;
}


//----------------------------------------------------------------------
//
// DriverEntry
//
// Installable driver initialization. Here we just set ourselves up.
//
//----------------------------------------------------------------------
NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath )
{
    NTSTATUS                ntStatus;
    PFILE_OBJECT            volumeObject;
    WCHAR                   deviceNameBuffer[]  = L"\\Device\\Undelete";
    UNICODE_STRING          deviceNameUnicodeString;
    WCHAR                   deviceLinkBuffer[]  = L"\\DosDevices\\Undelete";
    UNICODE_STRING          deviceLinkUnicodeString;
    WCHAR                   volumeBuffer[] = L"\\DosDevices\\X:\\";
    UNICODE_STRING          volumeBufferUnicodeString;
    PCHAR                   tmpFilterList;
    USHORT                  drive;
    ULONG                   i;
    PCHAR                   curDirPtr;
    HANDLE 		            volumeHandle;   
    OBJECT_ATTRIBUTES       objectAttributes;
    IO_STATUS_BLOCK  	    ioStatus;
    BOOLEAN                 errorMode;
    FILE_FS_DEVICE_INFORMATION fsDeviceInformation;
    RTL_QUERY_REGISTRY_TABLE QueryTable[4];    

    DbgPrint (("Undelete.SYS: entering DriverEntry\n"));

    //
    // If not NT 4.0 Final Release, shorten the Fast I/O table so that Undelete 
    // will work on the Betas and Release Candidates
    //
    if( *NtBuildNumber < NT4FINAL ) {

        FastIOHook.SizeOfFastIoDispatch = (ULONG) &FastIOHook.FastIoQueryNetworkOpenInfo - 
            (ULONG) &FastIOHook;
    } 

    //    
    // Setup the device name
    //    
    RtlInitUnicodeString (&deviceNameUnicodeString,
                          deviceNameBuffer );

    //
    // Create the device used for GUI communications
    //
    ntStatus = IoCreateDevice ( DriverObject,
                                0,
                                &deviceNameUnicodeString,
                                FILE_DEVICE_UNDELETE,
                                0,
                                TRUE,
                                &GUIDevice );

    //
    // If successful, make a symbolic link that allows for the device
    // object's access from Win32 programs
    //
    if (NT_SUCCESS(ntStatus)) {

        //
        // Create a symbolic link that the GUI can specify to gain access
        // to this driver/device
        //
        RtlInitUnicodeString (&deviceLinkUnicodeString,
                              deviceLinkBuffer );
        ntStatus = IoCreateSymbolicLink (&deviceLinkUnicodeString,
                                         &deviceNameUnicodeString );
        if (!NT_SUCCESS(ntStatus)) {

            DbgPrint (("Undelete.SYS: IoCreateSymbolicLink failed\n"));
            // FIX
            ntStatus = STATUS_SUCCESS;
        }

        //
        // Create dispatch points for all routines that must be handled. 
        // All entry points are registered since we might filter a
        // file system that processes all of them.
        //
        DriverObject->MajorFunction[IRP_MJ_READ]            = 
            DriverObject->MajorFunction[IRP_MJ_CLOSE]           =
            DriverObject->MajorFunction[IRP_MJ_WRITE]           =
            DriverObject->MajorFunction[IRP_MJ_CREATE]          =
            DriverObject->MajorFunction[IRP_MJ_FLUSH_BUFFERS]   =
            DriverObject->MajorFunction[IRP_MJ_QUERY_INFORMATION]   =
            DriverObject->MajorFunction[IRP_MJ_SET_INFORMATION] =
            DriverObject->MajorFunction[IRP_MJ_QUERY_EA]        =
            DriverObject->MajorFunction[IRP_MJ_SET_EA]          =
            DriverObject->MajorFunction[IRP_MJ_QUERY_VOLUME_INFORMATION] = 
            DriverObject->MajorFunction[IRP_MJ_SET_VOLUME_INFORMATION] =
            DriverObject->MajorFunction[IRP_MJ_DIRECTORY_CONTROL]   =
            DriverObject->MajorFunction[IRP_MJ_FILE_SYSTEM_CONTROL] =
            DriverObject->MajorFunction[IRP_MJ_QUERY_SECURITY]  =
            DriverObject->MajorFunction[IRP_MJ_SET_SECURITY]    =
            DriverObject->MajorFunction[IRP_MJ_SHUTDOWN]        =
            DriverObject->MajorFunction[IRP_MJ_LOCK_CONTROL]    =
            DriverObject->MajorFunction[IRP_MJ_CLEANUP]         =
            DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]  = UndeleteDispatch;

        //
        // Set up the Fast I/O dispatch table
        //
        DriverObject->FastIoDispatch = &FastIOHook;
    }

    //
    // If something went wrong, cleanup the device object and don't load
    //
    if (!NT_SUCCESS(ntStatus)) {

        DbgPrint(("Undelete: Failed to create our device!\n"));

        if( GUIDevice ) {

            IoDeleteDevice( GUIDevice );
        }
        return ntStatus;
    }

    //
    // Initialize and then read the filter
    //
    memset( QueryTable, 0, sizeof(RTL_QUERY_REGISTRY_TABLE) * 4);
    QueryTable[0].QueryRoutine = NULL;
    QueryTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT;
    QueryTable[0].EntryContext = &NumFilters;
    QueryTable[0].Name = L"NumFilters";
    QueryTable[0].DefaultType = REG_DWORD;
    QueryTable[0].DefaultData = &NumFilters;
    QueryTable[0].DefaultLength = sizeof(ULONG);

    QueryTable[1].QueryRoutine = NULL;
    QueryTable[1].Flags = RTL_QUERY_REGISTRY_DIRECT;
    QueryTable[1].EntryContext = &SizeDirFilters;
    QueryTable[1].Name = L"SizeDirFilters";
    QueryTable[1].DefaultType = REG_DWORD;
    QueryTable[1].DefaultData = &SizeDirFilters;
    QueryTable[1].DefaultLength = sizeof(ULONG);

    QueryTable[2].QueryRoutine = NULL;
    QueryTable[2].Flags = RTL_QUERY_REGISTRY_DIRECT;
    QueryTable[2].EntryContext = &NumDirFilters;
    QueryTable[2].Name = L"NumDirFilters";
    QueryTable[2].DefaultType = REG_DWORD;
    QueryTable[2].DefaultData = &NumDirFilters;
    QueryTable[2].DefaultLength = sizeof(ULONG);

    (void) RtlQueryRegistryValues( RTL_REGISTRY_ABSOLUTE, 
                                   L"\\Registry\\Machine\\SOFTWARE\\Systems Internals\\Undelete", 
                                   QueryTable, NULL, NULL );

    //
    // If there are filters, read them
    //
    if( NumFilters ) {

        //
        // Allocate a buffer with enough space for length and type which are returned at the front
        //
        FilterList = (PFILTERENTRY) ExAllocatePool( NonPagedPool, 
                                                    NumFilters * sizeof(FILTERENTRY));
        tmpFilterList = (PCHAR) ExAllocatePool( NonPagedPool, 
                                                NumFilters * sizeof(FILTERENTRY) + 2*sizeof(ULONG));
        *(ULONG *) tmpFilterList = NumFilters * sizeof(FILTERENTRY) + 2*sizeof(ULONG);

        memset( QueryTable, 0, sizeof(RTL_QUERY_REGISTRY_TABLE) * 2);
        QueryTable[0].QueryRoutine = NULL;
        QueryTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT;
        QueryTable[0].EntryContext = tmpFilterList;
        QueryTable[0].Name = L"Filter";
        QueryTable[0].DefaultType = REG_BINARY;
        QueryTable[0].DefaultData = tmpFilterList;
        QueryTable[0].DefaultLength = NumFilters * sizeof(FILTERENTRY);

        ntStatus = RtlQueryRegistryValues( RTL_REGISTRY_ABSOLUTE, 
                                           L"\\Registry\\Machine\\SOFTWARE\\Systems Internals\\Undelete", 
                                           QueryTable, NULL, NULL );
        if( !NT_SUCCESS( ntStatus )) {

            DbgPrint(("Undelete: Could not read %d filters: %x\n", NumFilters, ntStatus ));
            NumFilters = 0;
            ExFreePool( FilterList );
            FilterList = NULL;

        } else {

            DbgPrint(("Undelete: read %d filters\n", NumFilters ));
        }

        memcpy( FilterList, tmpFilterList + 2*sizeof(ULONG), NumFilters * sizeof(FILTERENTRY));
        ExFreePool( tmpFilterList );
    }

    //
    // If there are directory filters, read them
    //
    if( NumDirFilters ) {

        //
        // Allocate a buffer with enough space for length and type which are returned at the front
        //
        DirFilterBuffer = (PCHAR) ExAllocatePool( NonPagedPool, SizeDirFilters + 2*sizeof(ULONG));
        DirFilterArray = (PCHAR *) ExAllocatePool( NonPagedPool, NumDirFilters * sizeof(char *));

        if( !DirFilterArray || !DirFilterBuffer ) {

            DbgPrint(("Not enough pool to read in the directory filters.\n"));
            return STATUS_INSUFFICIENT_RESOURCES;
        }
      
        *(ULONG *) DirFilterBuffer = SizeDirFilters + 2*sizeof(ULONG);

        memset( QueryTable, 0, sizeof(RTL_QUERY_REGISTRY_TABLE) * 2);
        QueryTable[0].QueryRoutine = NULL;
        QueryTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT;
        QueryTable[0].EntryContext = DirFilterBuffer;
        QueryTable[0].Name = L"DirFilter";
        QueryTable[0].DefaultType = REG_BINARY;
        QueryTable[0].DefaultData = DirFilterBuffer;
        QueryTable[0].DefaultLength = 0;

        ntStatus = RtlQueryRegistryValues( RTL_REGISTRY_ABSOLUTE, 
                                           L"\\Registry\\Machine\\SOFTWARE\\Systems Internals\\Undelete", 
                                           QueryTable, NULL, NULL );
        if( !NT_SUCCESS( ntStatus )) {

            DbgPrint(("Undelete: Could not read %d filters: %x\n", NumFilters, ntStatus ));
            NumDirFilters = 0;
            ExFreePool( DirFilterArray );
            ExFreePool( DirFilterBuffer );
            DirFilterArray = NULL;
            DirFilterBuffer = NULL;
	    
        } else {

            DbgPrint(("Undelete: read %d directory filters\n", NumDirFilters ));
        }

        //
        // Now set up the the directory filter array
        //
        curDirPtr = DirFilterBuffer + 2*sizeof(ULONG);
        for( i = 0; i < NumDirFilters; i++ ) {

            DirFilterArray[i] = curDirPtr;
            curDirPtr += strlen( curDirPtr) + 1;
        }
    }

    //
    // Disable hard errors so that we don't pop-up when touching removable media
    //
    errorMode = IoSetThreadHardErrorMode( FALSE );

    //
    // Need to check each drive to see if its a local hard disk
    //
    RtlInitUnicodeString (&volumeBufferUnicodeString,
                          volumeBuffer );
    for(drive = 2; drive < 26; drive++ ) {

        LDriveDevices[drive] = NULL;

        volumeBufferUnicodeString.Buffer[12] = drive+'A';

        InitializeObjectAttributes( &objectAttributes, &volumeBufferUnicodeString, 
                                    OBJ_CASE_INSENSITIVE, NULL, NULL );

        ntStatus = ZwCreateFile( &volumeHandle, FILE_ANY_ACCESS, 
                                 &objectAttributes, &ioStatus, NULL, 0, FILE_SHARE_READ|FILE_SHARE_WRITE, 
                                 FILE_OPEN, 
                                 FILE_SYNCHRONOUS_IO_NONALERT|FILE_OPEN_FOR_BACKUP_INTENT|FILE_DIRECTORY_FILE, 
                                 NULL, 0 );
        if( !NT_SUCCESS( ntStatus ) ) {

            DbgPrint(("Undelete: Could not open drive %c: %x\n", 'A'+drive, ntStatus ));

        } else {

            //
            // Opened the volume. Now go and query it to see if its a local drive
            //
            ntStatus = ZwQueryVolumeInformationFile( volumeHandle, &ioStatus, &fsDeviceInformation,
                                                     sizeof(fsDeviceInformation), FileFsDeviceInformation );

            if( NT_SUCCESS( ntStatus )) {

                //
                // Alright, see if its the correct type 
                //
                if( fsDeviceInformation.DeviceType == FILE_DEVICE_DISK && !(fsDeviceInformation.Characteristics & 
                                                                            (FILE_REMOVABLE_MEDIA|FILE_READ_ONLY_DEVICE|FILE_FLOPPY_DISKETTE|FILE_WRITE_ONCE_MEDIA|
                                                                             FILE_REMOTE_DEVICE|FILE_VIRTUAL_VOLUME))) {
	                     
                    DbgPrint(("Volume %C must be filtered\n", 'A'+drive ));

                    //
                    // It is, so hook it.
                    //
                    HookDrive( drive, DriverObject, volumeHandle );
                }

                ZwClose( volumeHandle );
            }
        }
    }

    //
    // Save away system process
    //
    SystemProcess = PsGetCurrentProcess();

    IoSetThreadHardErrorMode( errorMode );

    return STATUS_SUCCESS;
}
